# Licensed under the Apache License: http://www.apache.org/licenses/LICENSE-2.0
# For details: https://github.com/gaogaotiantian/viztracer/blob/master/NOTICE.txt


from typing import TYPE_CHECKING


if TYPE_CHECKING:
    from .viztracer import VizTracer  # pragma: no cover


class VizEvent:
    def __init__(self, tracer: "VizTracer", event_name: str, file_name: str, lineno: int):
        self._tracer = tracer
        self.event_name = event_name
        self.file_name = file_name
        self.lineno = lineno
        self.start = None

    def __enter__(self):
        self.start = self._tracer.getts()

    def __exit__(self, type, value, trace):
        dur = self._tracer.getts() - self.start
        raw_data = {
            "ph": "X",
            "name": f"{self.event_name} ({self.file_name}:{self.lineno})",
            "ts": self.start,
            "dur": dur,
            "cat": "FEE"
        }
        self._tracer.add_raw(raw_data)



# Licensed under the Apache License: http://www.apache.org/licenses/LICENSE-2.0
# For details: https://github.com/gaogaotiantian/viztracer/blob/master/NOTICE.txt

import functools
from typing import Any, Callable, Dict, List, Optional
from .viztracer import VizTracer


class _EventBase:
    def __init__(self, tracer: VizTracer, name: str = "", **kwargs):
        self._viztracer_tracer: VizTracer = tracer
        self._viztracer_name: str = name
        self._viztracer_enable: bool = False
        self._viztracer_config: Dict = {
            "trigger_on_change": True,
            "include_attributes": [],
            "exclude_attributes": []
        }

        for key in kwargs:
            if key in self._viztracer_config:
                self._viztracer_config[key] = kwargs[key]

        self._viztracer_enable = True

    def __setattr__(self, name: str, value: Any) -> None:
        self.__dict__[name] = value
        if not name.startswith("_"):
            if self._viztracer_enable and self._viztracer_config["trigger_on_change"]:
                if self._viztracer_config["include_attributes"]:
                    if name in self._viztracer_config["include_attributes"]:
                        self._viztracer_log()
                elif self._viztracer_config["exclude_attributes"]:
                    if name not in self._viztracer_config["exclude_attributes"]:
                        self._viztracer_log()
                else:
                    self._viztracer_log()

    def _viztracer_get_attr_list(self) -> List[str]:
        if self._viztracer_config["include_attributes"]:
            return self._viztracer_config["include_attributes"]
        else:
            return [attr for attr in self.__dir__()
                    if not attr.startswith("_") and attr not in self._viztracer_config["exclude_attributes"]]

    def _viztracer_set_config(self, key: str, value: Any) -> None:
        if key not in self._viztracer_config:
            raise ValueError("No config named {}".format(key))
        self._viztracer_config[key] = value

    def config(self, key: str, value: Any) -> None:
        self._viztracer_set_config(key, value)

    def _viztracer_log(self) -> None:
        raise NotImplementedError("You should not use _EventBase class directly")

    def log(self) -> None:
        self._viztracer_log()

    @staticmethod
    def triggerlog(method: Optional[Callable] = None, when: str = "after"):
        if when not in ["after", "before", "both"]:
            raise ValueError("when has to be one of 'after', 'before' or 'both', not {}".format(when))

        def inner(func: Callable) -> Callable:

            @functools.wraps(func)
            def wrapper(self, *args, **kwargs) -> Any:
                if when == "before" or when == "both":
                    self._viztracer_log()
                ret = func(self, *args, **kwargs)
                if when == "after" or when == "both":
                    self._viztracer_log()
                return ret
            return wrapper

        if method:
            return inner(method)
        return inner



# Licensed under the Apache License: http://www.apache.org/licenses/LICENSE-2.0
# For details: https://github.com/gaogaotiantian/viztracer/blob/master/NOTICE.txt

import functools
from typing import Any, Callable, Optional
from viztracer import VizTracer, get_tracer
import os
import time


def ignore_function(method: Optional[Callable] = None, tracer: Optional[VizTracer] = None) -> Callable:

    def inner(func: Callable) -> Callable:

        @functools.wraps(func)
        def ignore_wrapper(*args, **kwargs) -> Any:
            # We need this to keep trace a local variable
            t = tracer
            if not t:
                t = get_tracer()
                if not t:
                    raise NameError("ignore_function only works with global tracer")
            t.pause()
            ret = func(*args, **kwargs)
            t.resume()
            return ret

        return ignore_wrapper

    if method:
        return inner(method)
    return inner


def trace_and_save(method: Optional[Callable] = None, output_dir: str = "./", **viztracer_kwargs):

    def inner(func: Callable) -> Callable:

        @functools.wraps(func)
        def wrapper(*args, **kwargs) -> Any:
            tracer = VizTracer(**viztracer_kwargs)
            tracer.start()
            ret = func(*args, **kwargs)
            tracer.stop()
            if not os.path.exists(output_dir):
                os.mkdir(output_dir)
            file_name = os.path.join(output_dir, "result_{}_{}.json".format(func.__name__, int(100000 * time.time())))
            tracer.fork_save(file_name)
            tracer.cleanup()
            return ret

        return wrapper

    if method:
        return inner(method)
    return inner


def log_sparse(func: Callable) -> Callable:
    tracer = get_tracer()
    if not tracer or not tracer.log_sparse:
        return func

    @functools.wraps(func)
    def wrapper(*args, **kwargs) -> Any:
        tracer = get_tracer()
        if tracer:
            # This should always be true
            start = tracer.getts()
            ret = func(*args, **kwargs)
            dur = tracer.getts() - start
            code = func.__code__
            raw_data = {
                "ph": "X",
                "name": f"{code.co_name} ({code.co_filename}:{code.co_firstlineno})",
                "ts": start,
                "dur": dur,
                "cat": "FEE"
            }
            tracer.add_raw(raw_data)
        else:  # pragma: no cover
            raise Exception("This should not be possible")
        return ret

    return wrapper



# Licensed under the Apache License: http://www.apache.org/licenses/LICENSE-2.0
# For details: https://github.com/gaogaotiantian/viztracer/blob/master/NOTICE.txt

try:
    import orjson as json  # type: ignore
except ImportError:
    import json  # type: ignore
import bisect
import os
from .functree import FuncTree
from .util import color_print
from . import __version__


class Frame:
    def __init__(self, parent, node):
        self.parent = parent
        self.node = node
        self.curr_children_idx = 0
        self.next = None
        self.code_string = None
        if parent:
            parent.next = self

    @property
    def curr_children_idx(self):
        return self.__curr_children_idx

    @curr_children_idx.setter
    def curr_children_idx(self, idx):
        self.code_string = None
        self.__curr_children_idx = idx

    def show(self, p):
        if self.code_string:
            p(self.code_string)
            return

        node = self.node
        firstlineno = node.lineno
        if node.filename:
            if not os.path.exists(node.filename):
                p("> Not able to locate the source")
                return

            with open(node.filename, encoding="utf-8") as f:
                lst = f.readlines()

            start = firstlineno - 1
            line = lst[start]
            stripped_line = line.lstrip()
            code_indent = len(line) - len(stripped_line)

            if node.funcname != "<module>":
                end = firstlineno
                indented = False
                while end < len(lst):
                    line = lst[end]
                    stripped_line = line.lstrip()
                    indent = len(line) - len(stripped_line)
                    if indent > code_indent:
                        indented = True
                    elif indented:
                        if len(stripped_line) > 0 and \
                                not stripped_line.startswith("#"):
                            break
                    end += 1

                # clear the ending spaces
                while lst[end - 1].strip() == "":
                    end -= 1
            else:
                end = len(lst)

            if self.curr_children_idx >= len(self.node.children):
                currline = end
            else:
                currline = self.node.children[self.curr_children_idx].caller_lineno - 1

            for idx in range(start, end):
                if len(lst[idx].strip()) > 0:
                    if idx == currline:
                        lst[idx] = "> " + lst[idx]
                    else:
                        lst[idx] = "  " + lst[idx]

            if currline == end:
                self.code_string = "".join(lst[start:end] + ["> \n"])
            else:
                self.code_string = "".join(lst[start:end] + ["  \n"])
        else:
            self.code_string = "> " + self.node.fullname
        p(self.code_string)


class CounterEvents:
    def __init__(self):
        self._events = []

    def add_event(self, event):
        # The events have to be added in order
        if self._events:
            args = self._events[-1]["args"].copy()
        else:
            args = {}
        args.update(event["args"])
        self._events.append({
            "ts": event["ts"],
            "args": args
        })

    def normalize(self, first_ts):
        for event in self._events:
            event["ts"] -= first_ts

    def get_args(self, ts):
        ts_lst = [event["ts"] for event in self._events]
        idx = bisect.bisect(ts_lst, ts)
        if idx == 0:
            return {}
        else:
            return self._events[idx - 1]["args"]


class ObjectEvents:
    def __init__(self):
        self._objects = {}

    def add_event(self, event):
        if event["id"] not in self._objects:
            entry = {
                "id": event["id"],
                "name": event["name"],
                "snapshots": [
                ]
            }
            self._objects[event["id"]] = entry

        entry = self._objects[event["id"]]

        if event["ph"] in ["N", "D"]:
            entry["snapshots"].append({
                "ts": event["ts"],
                "args": None
            })
        elif event["ph"] == "O":
            entry["snapshots"].append({
                "ts": event["ts"],
                "args": event["args"]["snapshot"]
            })
        else:  # pragma: no cover
            raise ValueError("Unexpected type for object event")

    def normalize(self, first_ts):
        for entry in self._objects.values():
            for snapshot in entry["snapshots"]:
                snapshot["ts"] -= first_ts

    def get_args(self, ts):
        ret = {}
        for entry in self._objects.values():
            ts_lst = [snapshot["ts"] for snapshot in entry["snapshots"]]
            idx = bisect.bisect(ts_lst, ts)
            if idx != 0:
                snapshot = entry["snapshots"][idx - 1]
                if snapshot["args"] is not None:
                    ret[entry["name"]] = snapshot["args"]
        return ret


class ProgSnapshot:
    compatible_version = "0.9.5"

    def __init__(self, json_string=None, p=print):
        self.func_trees = {}
        self.curr_node = None
        self.curr_frame = None
        self.first_tree = None
        self.curr_tree = None
        self.counter_events = CounterEvents()
        self.object_events = ObjectEvents()
        if json_string:
            if not self.load(json_string):
                self.valid = False
                return
        self.p = p
        self.valid = True

    def load(self, json_string):
        self.func_trees = {}
        raw_data = json.loads(json_string)
        if "viztracer_metadata" not in raw_data.keys():
            color_print("FAIL", "Error, json file version too old.")
            return False
        else:
            if not self.check_version(raw_data["viztracer_metadata"]["version"]):
                return False
        trace_events = raw_data["traceEvents"]
        for event in trace_events:
            self.load_event(event)
        self.first_tree = min([tree for tree in self.get_trees()], key=lambda x: x.first_ts())
        first_ts = self.first_tree.first_ts()
        self.curr_tree = self.first_tree
        self.curr_frame = Frame(None, self.first_tree.first_node())
        for tree in self.get_trees():
            tree.normalize(first_ts)
        self.counter_events.normalize(first_ts)
        self.object_events.normalize(first_ts)
        return True

    def load_event(self, event):
        # event is a chrome trace format event object
        try:
            ph = event["ph"]
            pid = event["pid"]
            tid = event["tid"]
        except Exception:
            raise ValueError("Error when loading event data: {}", event)

        if ph == "X":
            if pid not in self.func_trees:
                self.func_trees[pid] = {}
            if tid not in self.func_trees[pid]:
                self.func_trees[pid][tid] = FuncTree(pid, tid)

            self.func_trees[pid][tid].add_event(event)
        elif ph == "C":
            self.counter_events.add_event(event)
        elif ph in ["N", "D", "O"]:
            self.object_events.add_event(event)
        elif ph == "M":
            return
        else:
            print("Unsupported event type: {}".format(ph))
            return

    def check_version(self, version):
        def get_version_tuple(v):
            return tuple([int(elem) for elem in v.split(".")])
        if get_version_tuple(version) < get_version_tuple(self.compatible_version):
            color_print("FAIL", "Error, json file version too old.")
            return False
        elif get_version_tuple(version) > get_version_tuple(__version__):
            color_print("WARNING", "Warning, json file version is newer than vdb! Not sure if it's compatible")
        return True

    def get_trees(self):
        for pid in self.func_trees:
            for tid in self.func_trees[pid]:
                yield self.func_trees[pid][tid]
        return

    def show(self):
        self.curr_frame.show(self.p)

    def up(self):
        # Inspect previous frame
        if not self.curr_frame.parent:
            return False, "No outer frame anymore"
        self.curr_frame = self.curr_frame.parent
        return True, None

    def down(self):
        if not self.curr_frame.next:
            return False, "Already at current frame"
        self.curr_frame = self.curr_frame.next
        return True, None

    def _goto_inner_frame(self):
        while self.curr_frame.next:
            self.curr_frame = self.curr_frame.next

    def step(self):
        self._goto_inner_frame()
        curr_frame = self.curr_frame
        if curr_frame.curr_children_idx < len(curr_frame.node.children):
            child = curr_frame.node.children[curr_frame.curr_children_idx]
            new_frame = Frame(curr_frame, child)
            self.curr_frame = new_frame
        else:
            # go out of the function
            success, _ = self.func_return()
            if not success:
                return False, "at the end of the trace"
        return True, None

    def step_back(self):
        self._goto_inner_frame()
        curr_frame = self.curr_frame
        if curr_frame.curr_children_idx > 0:
            curr_frame.curr_children_idx -= 1
            child = curr_frame.node.children[curr_frame.curr_children_idx]
            new_frame = Frame(curr_frame, child)
            new_frame.curr_children_idx = len(new_frame.node.children)
            self.curr_frame = new_frame
        else:
            # go out of the function
            success, _ = self.func_return_back()
            if not success:
                return False, "at the beginning of the trace"
        return True, None

    def next(self):
        self._goto_inner_frame()
        curr_frame = self.curr_frame
        if curr_frame.curr_children_idx < len(curr_frame.node.children):
            curr_frame.curr_children_idx += 1
        else:
            # go out of the function
            success, _ = self.func_return()
            if not success:
                return False, "at the end of the trace"
        return True, None

    def next_back(self):
        self._goto_inner_frame()
        curr_frame = self.curr_frame
        if curr_frame.curr_children_idx > 0:
            curr_frame.curr_children_idx -= 1
        else:
            # go out of the function
            success, _ = self.func_return_back()
            if not success:
                return False, "at the beginning of the trace"
        return True, None

    def func_return(self):
        self._goto_inner_frame()
        curr_frame = self.curr_frame
        if not curr_frame.parent:
            # Check if there's a sibling
            node = curr_frame.node
            idx = node.parent.children.index(node)
            if idx < len(node.parent.children) - 1:
                self.curr_frame = Frame(None, node.parent.children[idx + 1])
                return True, None
            return False, "No callers available"
        curr_frame = curr_frame.parent
        curr_frame.curr_children_idx += 1
        curr_frame.next = None
        self.curr_frame = curr_frame

        return True, None

    def func_return_back(self):
        self._goto_inner_frame()
        curr_frame = self.curr_frame
        if not curr_frame.parent:
            # Check if there's a sibling
            node = curr_frame.node
            idx = node.parent.children.index(node)
            if idx > 0:
                self.curr_frame = Frame(None, node.parent.children[idx - 1])
                return True, None
            return False, "No callers available"
        curr_frame = curr_frame.parent
        curr_frame.next = None
        self.curr_frame = curr_frame

        return True, None

    def where(self):
        tmp_frame = self.curr_frame
        while tmp_frame.parent:
            tmp_frame = tmp_frame.parent
        while True:
            if tmp_frame == self.curr_frame:
                self.p("> " + tmp_frame.node.fullname)
            else:
                self.p("  " + tmp_frame.node.fullname)
            if tmp_frame.next:
                tmp_frame = tmp_frame.next
            else:
                break

        return True, None

    def goto_timestamp(self, ts):
        frame = Frame(None, self.curr_tree.node_by_timestamp(ts))
        while True:
            if frame.node.children:
                starts = [child.start for child in frame.node.children]
                idx = bisect.bisect_left(starts, ts)
                if idx == 0:
                    frame.curr_children_idx = 0
                    break
                else:
                    idx -= 1
                if frame.node.children[idx].end <= ts:
                    # here's what we need!
                    frame.curr_children_idx = idx + 1
                    break
                else:
                    new_frame = Frame(frame, frame.node.children[idx])
                    frame = new_frame
            else:
                break
        self.curr_frame = frame

        return True, None

    def get_timestamp(self):
        tmp_frame = self.curr_frame
        while tmp_frame.next:
            tmp_frame = tmp_frame.next
        if not tmp_frame.node.children:
            return tmp_frame.node.start
        if tmp_frame.curr_children_idx >= len(tmp_frame.node.children):
            return tmp_frame.node.children[-1].end
        else:
            return tmp_frame.node.children[tmp_frame.curr_children_idx].start

    def print_timestamp(self):
        self.p(str(self.get_timestamp()))

        return True, None

    def print_counter(self):
        self.p(str(self.counter_events.get_args(self.get_timestamp())))

        return True, None

    def print_object(self):
        object_dict = self.object_events.get_args(self.get_timestamp())
        for k, v in object_dict.items():
            self.p(str(k) + ":")
            self.p(str(v))

        return True, None

    def list_tid(self):
        curr_tree = self.curr_tree
        forest = self.func_trees[curr_tree.pid]
        for tid in forest:
            if tid == curr_tree.tid:
                self.p("> {}".format(tid))
            else:
                self.p("  {}".format(tid))

        return True, None

    def list_pid(self):
        curr_tree = self.curr_tree
        for pid in self.func_trees:
            if pid == curr_tree.pid:
                self.p("> {}".format(pid))
            else:
                self.p("  {}".format(pid))

        return True, None

    def goto_tid(self, tid):
        assert(type(tid) is int)
        forest = self.func_trees[self.curr_tree.pid]
        if tid not in forest:
            return False, "No such tid"
        else:
            ts = self.get_timestamp()
            self.curr_tree = forest[tid]
            self.goto_timestamp(ts)

        return True, None

    def goto_pid(self, pid):
        assert(type(pid) is int)
        if pid not in self.func_trees:
            return False, "No such pid"
        else:
            ts = self.get_timestamp()
            forest = self.func_trees[pid]
            for tid in forest:
                self.curr_tree = forest[tid]
                break
            self.goto_timestamp(ts)

        return True, None

    def print_args(self):
        self.p(str(self.curr_frame.node.event.get("args", "")))

        return True, None



# Licensed under the Apache License: http://www.apache.org/licenses/LICENSE-2.0
# For details: https://github.com/gaogaotiantian/viztracer/blob/master/NOTICE.txt

import sys
import os
import argparse
try:
    import readline  # noqa: F401
except ImportError:
    pass

from .prog_snapshot import ProgSnapshot


class Simulator:
    def __init__(self, json_string, no_clear=False, extra_newline=False):
        try:
            from rich.console import Console  # type: ignore
            from rich.syntax import Syntax  # type: ignore
            self.console = Console()

            def p(s):
                syntax = Syntax(s, "python", theme="monokai")
                self.console.print(syntax, soft_wrap=True)

            self.print = p
        except ImportError:
            self.print = print
        self.snapshot = ProgSnapshot(json_string, self.print)
        if not self.snapshot.valid:
            exit(1)
        self.no_clear = no_clear
        self.extra_newline = extra_newline

    def start(self):
        self.clear()
        self.snapshot.show()
        while True:
            try:
                cmd = input(">>> ")
                if self.extra_newline:
                    print("")
                self.parse_cmd(cmd)
            except EOFError:
                exit(0)

    def clear(self):
        if not self.no_clear:
            os.system("cls" if os.name == "nt" else "clear")

    def parse_cmd(self, cmd):
        args = cmd.split()
        if not args:
            return
        success = False
        if args[0] == "s":
            success, err_msg = self.snapshot.step()
        elif args[0] == "sb":
            success, err_msg = self.snapshot.step_back()
        elif args[0] == "n":
            success, err_msg = self.snapshot.next()
        elif args[0] == "nb":
            success, err_msg = self.snapshot.next_back()
        elif args[0] == "r":
            success, err_msg = self.snapshot.func_return()
        elif args[0] == "rb":
            success, err_msg = self.snapshot.func_return_back()
        elif args[0] == "t":
            if len(args) == 1:
                success, err_msg = self.snapshot.print_timestamp()
                return
            elif len(args) == 2:
                success, err_msg = self.snapshot.goto_timestamp(float(args[1]))
            else:
                success, err_msg = False, "t takes 1 or no argument"
        elif args[0] == "u":
            success, err_msg = self.snapshot.up()
        elif args[0] == "d":
            success, err_msg = self.snapshot.down()
        elif args[0] == "w":
            success, err_msg = self.snapshot.where()
            return
        elif args[0] == "tid":
            if len(args) == 1:
                success, err_msg = self.snapshot.list_tid()
                return
            elif len(args) == 2:
                try:
                    tid = int(args[1])
                    success, err_msg = self.snapshot.goto_tid(tid)
                except ValueError:
                    print("tid needs to be an integer")
                    return
            else:
                success, err_msg = False, "tid takes 1 or no argument"
        elif args[0] == "pid":
            if len(args) == 1:
                success, err_msg = self.snapshot.list_pid()
                return
            elif len(args) == 2:
                try:
                    pid = int(args[1])
                    success, err_msg = self.snapshot.goto_pid(pid)
                except ValueError:
                    print("tid needs to be an integer")
                    return
            else:
                success, err_msg = False, "pid takes 1 or no argument"
        elif args[0] in ["arg", "args"]:
            success, err_msg = self.snapshot.print_args()
            return
        elif args[0] == "counter":
            success, err_msg = self.snapshot.print_counter()
            return
        elif args[0] == "object":
            success, err_msg = self.snapshot.print_object()
            return
        elif args[0] in ["quit", "exit", "q"]:
            exit(0)
        else:
            print("Invalid command: {}".format(cmd))
            return

        if success:
            self.clear()
            self.snapshot.show()
        else:
            print(err_msg)


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("file", nargs=1)
    parser.add_argument("--no_clear", action="store_true", default=False)
    parser.add_argument("--extra_newline", action="store_true", default=False)

    options = parser.parse_args(sys.argv[1:])

    filename = options.file[0]
    with open(filename, encoding="utf-8") as f:
        s = f.read()

    sim = Simulator(s, no_clear=options.no_clear, extra_newline=options.extra_newline)
    sim.start()


if __name__ == '__main__':
    main()



# Licensed under the Apache License: http://www.apache.org/licenses/LICENSE-2.0
# For details: https://github.com/gaogaotiantian/viztracer/blob/master/NOTICE.txt

from logging import Handler, LogRecord
from .viztracer import VizTracer


class VizLoggingHandler(Handler):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._tracer = None

    def emit(self, record: LogRecord):
        if not self._tracer:
            return
        self._tracer.add_instant(f"logging - {self.format(record)}", scope="p")

    def setTracer(self, tracer: VizTracer):
        self._tracer = tracer



# Licensed under the Apache License: http://www.apache.org/licenses/LICENSE-2.0
# For details: https://github.com/gaogaotiantian/viztracer/blob/master/NOTICE.txt

from string import Template
import os
import re
import gzip
import sys
from typing import Any, Dict, List, Sequence, Union, TextIO
try:
    import orjson  # type: ignore
except ImportError:
    import json
from .util import color_print


def get_json(data: Union[Dict, str]) -> Dict[str, Any]:
    # This function will return a json object if data is already json object
    # or a opened file or a file path
    if isinstance(data, dict):
        # This is an object already
        return data
    elif isinstance(data, str):
        with open(data, encoding="utf-8") as f:
            json_str = f.read()
    else:
        raise TypeError("Unexpected Type{}!", type(data))

    try:
        if "orjson" in sys.modules:
            return orjson.loads(json_str)
        else:
            return json.loads(json_str)
    except Exception as e:
        print("Unable to decode {}".format(data))
        raise e


class ReportBuilder:
    def __init__(
            self,
            data: Union[Sequence[str], Dict],
            verbose: int = 1,
            align: bool = False,
            minimize_memory: bool = False):
        self.verbose = verbose
        self.combined_json: Dict = {}
        self.entry_number_threshold = 4000000
        self.align = align
        self.minimize_memory = minimize_memory
        if isinstance(data, dict):
            self.jsons = [get_json(data)]
        elif isinstance(data, (list, tuple)):
            self.jsons = [get_json(j) for j in data]
        else:
            raise TypeError("Invalid data type for ReportBuilder")

    def combine_json(self) -> None:
        if self.combined_json:
            return
        if not self.jsons:
            raise ValueError("Can't get report of nothing")
        if self.align:
            for one in self.jsons:
                self.align_events(one["traceEvents"])
        self.combined_json = self.jsons[0]
        for one in self.jsons[1:]:
            if "traceEvents" in one:
                self.combined_json["traceEvents"].extend(one["traceEvents"])

    def align_events(self, original_events: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
        """
        Apply an offset to all the trace events, making the start timestamp 0
        This is useful when comparing multiple runs of the same script

        This function will change the timestamp in place, and return the original list
        """
        offset_ts = min((event["ts"] for event in original_events if "ts" in event))
        for event in original_events:
            if "ts" in event:
                event["ts"] -= offset_ts
        return original_events

    def prepare_json(self, file_info: bool = True, display_time_unit: str = "us") -> None:
        # This will prepare self.combined_json to be ready to output
        self.combine_json()
        if self.verbose > 0:
            entries = len(self.combined_json["traceEvents"])
            print(f"Dumping trace data, total entries: {entries}")

        self.combined_json["displayTimeUnit"] = display_time_unit

        if file_info:
            self.combined_json["file_info"] = {"files": {}, "functions": {}}
            pattern = re.compile(r".*\((.*):([0-9]*)\)")
            file_dict = self.combined_json["file_info"]["files"]
            func_dict = self.combined_json["file_info"]["functions"]
            for event in self.combined_json["traceEvents"]:
                if event["ph"] == 'X':
                    if event["name"] not in func_dict:
                        try:
                            m = pattern.match(event["name"])
                            if m is not None:
                                file_name = m.group(1)
                                lineno = int(m.group(2))
                                if file_name not in file_dict:
                                    with open(file_name, "r", encoding="utf-8") as f:
                                        content = f.read()
                                        file_dict[file_name] = [content, content.count("\n")]
                                func_dict[event["name"]] = [file_name, lineno]
                        except Exception:
                            pass

    def generate_report(
            self,
            output_file: TextIO,
            output_format: str,
            file_info: bool = True) -> None:
        sub = {}
        if output_format == "html":
            self.prepare_json(file_info=file_info, display_time_unit="ns")
            with open(os.path.join(os.path.dirname(__file__), "html/trace_viewer_embedder.html"), encoding="utf-8") as f:
                tmpl = f.read()
            with open(os.path.join(os.path.dirname(__file__), "html/trace_viewer_full.html"), encoding="utf-8") as f:
                sub["trace_viewer_full"] = f.read()
            if "orjson" in sys.modules:
                sub["json_data"] = orjson.dumps(self.combined_json) \
                                         .decode("utf-8") \
                                         .replace("</script>", "<\\/script>")
            else:
                sub["json_data"] = json.dumps(self.combined_json) \
                                       .replace("</script>", "<\\/script>")
            output_file.write(Template(tmpl).substitute(sub))
        elif output_format == "json":
            self.prepare_json(file_info=file_info, display_time_unit="us")
            if "orjson" in sys.modules:
                output_file.write(orjson.dumps(self.combined_json).decode("utf-8"))
            else:
                if self.minimize_memory:
                    json.dump(self.combined_json, output_file)  # type: ignore
                else:
                    output_file.write(json.dumps(self.combined_json))

    def save(self, output_file: Union[str, TextIO] = "result.html", file_info: bool = True) -> None:
        if self.verbose > 0:
            print("==================================================")
            print("== Starting from version 0.13.0, VizTracer will ==")
            print("== use json as the default report file. You can ==")
            print('== generate HTML report with "-o result.html"   ==')
            print("==================================================")

        if isinstance(output_file, str):
            file_type = output_file.split(".")[-1]

            if file_type == "html":
                with open(output_file, "w", encoding="utf-8") as f:
                    self.generate_report(f, output_format="html", file_info=file_info)
            elif file_type == "json":
                with open(output_file, "w", encoding="utf-8") as f:
                    self.generate_report(f, output_format="json", file_info=file_info)
            elif file_type == "gz":
                with gzip.open(output_file, "wt") as f:
                    self.generate_report(f, output_format="json", file_info=file_info)
            else:
                raise Exception("Only html, json and gz are supported")
        else:
            self.generate_report(output_file, output_format="json", file_info=file_info)

        if self.verbose > 0 and isinstance(output_file, str):
            print("Saving report to {} ...".format(os.path.abspath(output_file)))
            print('Use', end=" ")
            color_print("OKGREEN", '"vizviewer <your_report>"', end=" ")
            print('to open the report')



# Licensed under the Apache License: http://www.apache.org/licenses/LICENSE-2.0
# For details: https://github.com/gaogaotiantian/viztracer/blob/master/NOTICE.txt

from .event_base import _EventBase


class VizCounter(_EventBase):
    def _viztracer_log(self):
        if not self._viztracer_tracer:
            return
        d = {}
        for attr in self._viztracer_get_attr_list():
            if hasattr(self, attr):
                val = self.__getattribute__(attr)
                if not callable(val):
                    if type(val) is int or type(val) is float:
                        d[attr] = val
                    else:
                        raise Exception("Counter can only take numeric values")
        self._viztracer_tracer.add_counter(self._viztracer_name, d)



# Licensed under the Apache License: http://www.apache.org/licenses/LICENSE-2.0
# For details: https://github.com/gaogaotiantian/viztracer/blob/master/NOTICE.txt

try:
    from IPython.core.magic import cell_magic, magics_class, Magics, needs_local_scope  # type: ignore

    @magics_class
    class VizTracerMagics(Magics):
        @needs_local_scope
        @cell_magic
        def viztracer(self, line, cell, local_ns):
            from .viztracer import VizTracer
            from .viewer import view
            from IPython.display import display  # type: ignore
            from ipywidgets import Button  # type: ignore
            code = self.shell.transform_cell(cell)
            file_path = "./viztracer_report.json"
            with VizTracer(verbose=0, output_file=file_path):
                exec(code, local_ns, local_ns)

            button = Button(description="VizTracer Report")
            button.on_click(lambda b: view(file_path, once=True))

            display(button)

except ImportError:  # pragma: no cover
    pass


def load_ipython_extension(ipython):
    """
    Use `%load_ext viztracer`
    """
    ipython.register_magics(VizTracerMagics)



# Licensed under the Apache License: http://www.apache.org/licenses/LICENSE-2.0
# For details: https://github.com/gaogaotiantian/viztracer/blob/master/NOTICE.txt

from .main import main


if __name__ == '__main__':
    main()
